/*
  (c) 2008 Ender
  Patch utility (meant for Phoenix BIOS modules)
*/

#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>

void ensure_final_slash(char *str)
{
    short len = strlen(str);
    if(str[len - 1] != '/') {
        str[len] = '/';
        str[len + 1] = 0;
    }
}

int strendcmp(const char *a, const char *b)
{
    return strcmp(a + strlen(a) - strlen(b), b);
}

#define ERR_CANNOT_OPEN -1
#define ERR_CANNOT_READ -2
int load_file(char *fn, char **buf)
{
    FILE *f;
    int size;
    
	if(!(f = fopen(fn, "rb"))) {
	    return ERR_CANNOT_OPEN;
	}

    fseek(f, 0, SEEK_END);
    size = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    if(size == 0) {
        *buf = NULL;
        return 0;
    }
	
	*buf = (char *) malloc(size);
	if(fread(*buf, size, 1, f) != 1) {
	    free(*buf);
	    return ERR_CANNOT_READ;
	}
	fclose(f);
	
	return size;
}

void memand(char *a, char *b, int len)
{
    while(len--) {
       *a = (*a) & (*b);
       a++;
       b++;
    }
}

#define OP_OVERWRITE 1
short apply_patch(char *buf, char *patch, int pchsize)
{
    int k;
    short operation, pos, len;
    
    for(k = 0; k < pchsize; ) {
        operation = *((short *) (patch + k));
        
        switch(operation) {
            case OP_OVERWRITE:
                pos = *((short *) (patch + k + 2));
                len = *((short *) (patch + k + 4));
                memcpy(buf + pos, patch + k + 6, len);
                
                k += 6 + len;
                break;
            
            default:
                printf("Unsupported patch operation. Data in inconsistent state, DO NOT USE.\n");
                return 0;
        }
    }
    
    return 1;
}

int find_and_apply_patch(char *buf, int size, char *pattern, char *mask, int patsize, char *patch, int pchsize)
{
    int i, matches = 0;
    char *testbuf, *bufend = buf + size - patsize;
    
    if(size < patsize) return 0;

    testbuf = (char *) malloc(patsize);
    for(; buf != bufend; buf++) {
        memcpy(testbuf, buf, patsize);
        memand(testbuf, mask, patsize);
        if(memcmp(testbuf, pattern, patsize) == 0) {
            matches++;
            if(!apply_patch(buf, patch, pchsize)) {
                printf("Application of patch failed. Terminating.\n");
                exit(5);
            }
        }
    }
    free(testbuf);
    return matches;
}

char *memsearch(char *haystack, char *needle, int haystack_len, int needle_len)
{
    char *match_begin = haystack, *nd_end = needle + needle_len, *hs_end = haystack + haystack_len;
    char *hs = haystack, *nd = needle;
    while(hs != hs_end) {
       if(*nd != *hs) {
           nd = needle;
           hs = ++match_begin;
       }
       nd++; hs++;
       if(nd == nd_end) {
           return match_begin;
       }
    }
    return NULL;
}

int main(int argc, char **argv)
{
    DIR *dh;
    struct dirent *fn;
    struct stat fileinfo;
    char startdir[256], moduledir[512], patchesdir[512], filename[512];
    char *buf;
    char *patterns[256], *masks[256], *patches[256], patchnames[256][128];
    int patternsize[256], patchsize[256], numpatches = 0;
    int size, i, totalnum;
    FILE *f;

    printf("PhnxPatch 0.1 2008-03-02\n(c) Ender\n\n");
    if(argc != 3) {
        printf("This utility will patch module(s) of Phoenix BIOS.\n"
               "Usage: \n"
               "  phnxpatch patches bios_modules\n"
               "\n"
               "Where 'bios_modules' is a path to directory with extracted BIOS\n"
               "modules (all files in there will be searched and patched if needed)\n"
               "and 'patches' is a path to directory with patch files (see docs).\n");
        return 0;
    }

    strcpy(patchesdir, argv[1]);
    ensure_final_slash(patchesdir);
    strcpy(moduledir, argv[2]);
    ensure_final_slash(moduledir);
    
    if(!(dh = opendir(patchesdir))) {
        printf("Failed to open patches directory (%s).\n", patchesdir);
        return 5;
    }

    printf("Scanning patches directory %s...\n", patchesdir);
    while(fn = readdir(dh)) {
        if(strendcmp(fn->d_name, ".pattern") != 0) {
            continue;
        }

        strcpy(patchnames[numpatches], fn->d_name);
        strrchr(patchnames[numpatches], '.')[0] = 0;
        printf("  %s:\n", patchnames[numpatches]);

        /* Read pattern */
        strcpy(filename, patchesdir);
        strcat(filename, fn->d_name);
        size = load_file(filename, patterns + numpatches);
        if(size == ERR_CANNOT_OPEN || size == ERR_CANNOT_READ) {
            printf("    failed to open or read, skipping.\n");
            continue;
        }
        patternsize[numpatches] = size;
        printf("    pattern read\n");
        
        /* Read mask */
        strcpy(strrchr(filename, '.'), ".mask");
        size = load_file(filename, masks + numpatches);
        if(size == ERR_CANNOT_OPEN || size == ERR_CANNOT_READ) {
            free(patterns[numpatches]);
            printf("    failed to open or read mask, skipping.\n");
            continue;
        }
        else if(size != patternsize[numpatches]) {
            free(patterns[numpatches]);
            free(masks[numpatches]);
            printf("    mask size not equal to pattern, skipping.\n");
            continue;
        }
        printf("    mask read\n");
        
        /* Read patch */
        strcpy(strrchr(filename, '.'), ".patch");
        size = load_file(filename, patches + numpatches);
        if(size == ERR_CANNOT_OPEN || size == ERR_CANNOT_READ) {
            free(patterns[numpatches]);
            free(masks[numpatches]);
            printf("    failed to open or read patch, skipping.\n");
            continue;
        }
        patchsize[numpatches] = size;
        printf("    patch read\n");
        numpatches++;

        printf("    done.\n");
    }
    printf("\n");

    if(!(dh = opendir(moduledir))) {
        printf("Failed to open module directory (%s).\n", moduledir);
        return 5;
    }

    printf("Scanning modules directory %s...\n", moduledir);
    while(fn = readdir(dh)) {
        strcpy(filename, moduledir);
        strcat(filename, fn->d_name);
        if(strcmp(fn->d_name, ".") == 0 || strcmp(fn->d_name, "..") == 0) {
            continue;
        }
        if(strcmp(strrchr(fn->d_name, '.'), ".patched") == 0) {
            continue;
        }
        
        printf("%s: ", fn->d_name);

        stat(filename, &fileinfo);
        if((fileinfo.st_mode & S_IFDIR) != 0) {
            printf("is a directory, skipping.\n", fn->d_name);
            continue;
        }
        
        if(!(f = fopen(filename, "rb"))) {
            printf("failed to open, skipping.\n");
            continue;
        }
        printf("\n");
        fseek(f, 0, SEEK_END);
        size = ftell(f);
        fseek(f, 0, SEEK_SET);

        /* read whole module to buffer */
        buf = (char *) malloc(size);
        fread(buf, size, 1, f);
        fclose(f);
        
        /* try apply all patches one after another */
        totalnum = 0;
        for(i = 0; i < numpatches; i++) {
            int n = find_and_apply_patch(buf, size, patterns[i], masks[i], patternsize[i], patches[i], patchsize[i]);
            if(n) {
                printf("  patch %s matched and applied %d time(s);\n", patchnames[i], n);
            }
            else {
                printf("  patch %s not matched, not applied;\n", patchnames[i]);
            }
            totalnum += n;
        }

        /* if modified, write result */
        if(totalnum > 0) {
            strcat(filename, ".patched");
            if(!(f = fopen(filename, "wb")) ||
               fwrite(buf, size, 1, f) != 1 ||
               fclose(f) == EOF) {
                free(buf);
                printf("failed to open or write output file, skipping.\n");
                continue;
            }
            printf("  wrote changed file as %s;\n", filename);
        }
        
        printf("  done.\n");
    }
        
    printf("\n"
           "---------------- WARNING! ----------------\n");
    printf("If you don't know what you're doing there's a very high risk of rendering your\n"
           "computer unusable by flashing a modified BIOS. This is not a safe playground.\n");
    printf("---------------- WARNING! ----------------\n");
    
    return 0;
}
